#!/usr/bin/perl
##############################################################################
#
#  ETS -- Wolfenstein: Enemy Territory Statistics
#  (c) Vladi Shabanski <cade@bis.bg> 2003
#  http://soul.datamax.bg/~cade/away/ets
#
#  Developed for play.evrocom.net server ET community
#  http://play.evrocom.net
#
#  This software is distributed under GPL license, see COPYING for details.
#
#  $Id: ets.pl,v 1.77 2004/03/01 02:22:58 cade Exp $
#
##############################################################################
use Data::Dumper;
use strict;

### GET PARAMETERS ###########################################################

my $help = $ARGV[0];

my $log_file     = shift || $ENV{ 'ETS_LOG_FILE' };
my $log_con_file =          $ENV{ 'ETS_CON_FILE' };
my $log_pb_file  =          $ENV{ 'ETS_PB_FILE' };
my $output_dir   = shift || $ENV{ 'ETS_OUTPUT_DIR' };
my $web_prefix   =          $ENV{ 'ETS_WEB_PREFIX' };
my $web_dir      =          $ENV{ 'ETS_WEB_DIR' };
my $theme        = shift || $ENV{ 'ETS_THEME' } || "ets-default";
my $last_days    =          $ENV{ 'ETS_LAST_DAYS' };
my $motd_file    =          $ENV{ 'ETS_MOTD_FILE' };

print "CONFIG:
  log_file     = $log_file  
  output_dir   = $output_dir
  web_prefix   = $web_prefix
  web_dir      = $web_dir   
  theme        = $theme     
";

die "usage1: $0 log-file html-output-dir [theme-name]
usage2: . config && $0
"  if $help =~ /^--?h(elp)?/i or $log_file eq '';

### CONFIG ###################################################################

our $FLOOR = 50;

our $TOPS_FLOOR = 15;

our $SCORE_FLOOR = 100;

our %TC;
our $T;
our $PROFILE = 0;

# files_dir is the web path of the dir where the web-content should be written to
# my $files_dir = "/mk/gaming/stats/et";
#my $files_dir = "/etstats";

# Sascha:
# this dir is the one you use to call the referred pages from. There is no difference for
# $web_dir and $file_dir as long you do _not_ run in a postnuke (or considerable) content.
# Be sure you have htmlloader.php or a similar Content-Caller like PostWrap installed if
# you use this calling method!
#
# my $web_prefix = "/postnuke/html/htmlloader.php?content=" . $files_dir;
#my $web_prefix = $files_dir;

our $revision = '$Revision: 1.77 $';
our $ets_version = $1 if $revision =~ /([\d\.]+)/;

### CONFIG ###################################################################

my $html_header = "<html>
        <meta http-equiv='content-type' content='text/html; charset=iso-8859-1'>
        <title>Wolfenstein: Enemy Territory Statistics</title>
				<link href=#$theme/styles/ets-default.css rel=stylesheet type=text/css>
				<style type=text/css></style>
        <center>
        <p>
        <h1>Wolfenstein: Enemy Territory Statistics</h1>
        <h2>[ENV:ETS_SERVER]</h2>
        <small>[T:This page has been generated on] [TIME]</small>
        <p>
        |
        <a href=!tops.html>[T:top scores]</a> |
        <a href=!games.html>[T:last games]</a> |
        <a href=!players.html>[T:players (long!)]</a> |
        <a href=!clans.html>[T:clans]</a> |
        <a href=!classide.html>[T:sides/classes]</a> |
        <br>
        |
        <a href=!wep.html>[T:weapons]</a> |
        <a href=!covert.html>[T:covert operations]</a> |
        <a href=!kpm.html>[T:top kills per minute]</a> |
        <a href=!noise.html>[T:noise]</a> |
        <a href=!waste.html>[T:time waste]</a> |
        <a href=!punks.html>[T:hall of shame]</a> |
        <br>
        |
        <a href=!p_sides.html>[T:sides played]</a> |
        <a href=!top_td.html>[T:'no friends' ladder]</a> |
        <a href=!harvest.html>[T:'harvest' ladder]</a> |
        <p>
        <table width=80%><td valign=top align=center>
        ";

my $html_footer = "</td></table><p><small>";
$html_footer .= "Note: Statistics data for last $last_days days only is used.<br>" if $last_days > 0;
$html_footer .= "Note: Limit is set to top $FLOOR entries (the rest are counted but not shown).<br>
       Thanks to all players at <a href=http://play.evrocom.net>http://play.evrocom.net</a>!<br>
       <a href=http://soul.datamax.bg/~cade/away/ets><b>ETS $ets_version</b></a> by (c) <a href=mailto:cade\@bis.bg>&lt;cade\@bis.bg&gt;</a> 2003
       </small>
       </center>
       </html>";

my $html_motd =               $ENV{ 'ETS_MOTD' }
                || load_file( $ENV{ 'ETS_MOTD_FILE' } )
                || "<h2>Welcome to Wolfenstein: Enemy Territory server statistics!</h2>";

###STATIC#####################################################################

my %CLASS = (
             0 => 'SOLDIER',
             1 => 'MEDIC',
             2 => 'ENGINEER',
             3 => 'FIELD_OPS',
             4 => 'COVERT_OPS',
            );
            
my %SIDE  = (
             0 => 'UNKNOWN',
             1 => 'AXIS',
             2 => 'ALLIES',
             3 => 'SPECTATORS',
            );            
my %WEAPON = (
             0 => 'KNIFE',
             1 => 'LUGER',
             2 => 'COLT',             #  8
             3 => 'MP40',             #  9
             4 => 'THOMPSON',         # 10
             5 => 'STEN',             # 11
             6 => 'FG42',             # 15
             7 => 'PANZERFAUST',
             8 => 'FLAMETHROWER',
             9 => 'GRENADE',
            10 => 'MORTAR',
            11 => 'DYNAMITE',         # 26
            12 => 'AIRSTRIKE',        # 28
            13 => 'ARTY',             # 30
            14 => 'SYRINGE',
            15 => 'SMOKECAN',
            16 => 'SATCHEL',
            17 => 'GRENADE_LAUNCHER', # 18
            18 => 'LANDMINE',         # 45
            19 => 'MOBILE_MG42',
            20 => 'GARAND',
            21 => 'KAR98'
            );
my %WEAPON_REV = reverse %WEAPON;            

my %SKILL = (
             0 => 'BATTLE SENSE',             
             1 => 'ENGINEERING   (ENGINEER)',             
             2 => 'FIRST AID     (MEDIC)',
             3 => 'SIGNALS       (FIELD OPS)',
             4 => 'LIGHT WEAPONS',
             5 => 'HEAVY WEAPONS (SOLDIER)',
             6 => 'COVERT OPS    (COVERT OPS)',
            );
my %SKILL_IMG = (
             0 => "<img src=#$theme/img/bs.gif border=0>",
             1 => "<img src=#$theme/img/en.gif border=0>",
             2 => "<img src=#$theme/img/fa.gif border=0>",
             3 => "<img src=#$theme/img/si.gif border=0>",
             4 => "<img src=#$theme/img/lw.gif border=0>",
             5 => "<img src=#$theme/img/hw.gif border=0>",
             6 => "<img src=#$theme/img/co.gif border=0>",
            );
            
# thanx, furiozo :)
my %COLOR = (
            '1' => 'f70400',
            '2' => '02f702',
            '3' => 'fdf900',
            '4' => '0006f8',
            '5' => '00f6f3',
            '6' => 'f407f5',
            '7' => 'f9f9f9',
            '8' => 'ff7b06',
            '9' => '7c7d81',
            '0' => '000100',
            'q' => 'f70303',
            ')' => 'f70303',
            'w' => 'f4f6f1',
            'e' => '8100fd',
            'r' => '00f904',
            't' => '0005f4',
            'y' => '7b7d7a',
            'u' => '00faf6',
            'i' => 'fb0033',
            'o' => 'eff174',
            'p' => '060702',
            'a' => 'fd9534',
            's' => 'fbf904',
            'd' => '007dfe',
            'f' => '3795c8',
            '&' => '3795c8',
            'g' => 'c2fcc9',
            'h' => '066432',
            '(' => '066432',
            'j' => 'ad1c17',
            'k' => '9b3107',
            'l' => 'cf9834',
            ';' => 'c0bfc7',
            'z' => 'bcbdc2',
            'x' => 'f98000',
            'c' => '7b0484',
            'v' => 'f900f9',
            'b' => '037d7e',
            'n' => 'faf9c0',
            'm' => '9b9935',
            ',' => 'cf9a3a',
            '.' => 'f7f8b6',
            '/' => 'fbfc7c',
            '-' => '98972d',
            '=' => '7f7f00',
            '\\' => '007f03',
            '[' => 'bec0bb',
            ']' => '7f7c0b',
            '\'' => 'c5f9c8',
            '*' => 'ae181a',
            );

###MAIN#######################################################################

my %name_map = split /(?<!\\)[=\n]+/, `cat $ENV{ETS_NAME_MAP}`
  if -e $ENV{ETS_NAME_MAP};

my @clan_map = split /(?<!\\)[=\n]+/, `cat $ENV{ETS_CLAN_MAP}`
  if -e $ENV{ETS_CLAN_MAP};
s/\\(.)/$1/g for @clan_map;
my %clan_map = @clan_map;  

my %translation = split /(?<!\\)[=\n]+/, `cat $ENV{ETS_TRANSLATION}`
  if -e $ENV{ETS_TRANSLATION};

my %kil; # koi-kogo e kill-nal
my %ded; # ...i obratnoto

my %nam;
my %nam_color;

my %unif_taken; # uniforms taken
my %unif_lost; # uniforms lost

my %noise; # noise in words :)

my %class_last; # last classes per player
my %class; # classes per player

my %side_last; # last side per player
my %side; # side per player

my %global_class;
my %global_side;

my %classes;
my %sides;

my %client; # clients info

my $href = 0;
my %href;

my $minute; # used for top kills per minute
my %kpm_temp;
my %kpm_game; # which minute in which game is
my %kpm;

my %top_td;     # team damage

my %conn_time;  # connect time
my %time_waste; # time "wasted" :)

my @games; # all games
my %games; # game time->game data
my @trans; # last game transcription

my %player; # stats from weapon stats, same as in @games

my %weapons; # stats total
my %skills;  # skills total

my %top_snipe;   # top usage of sniper weapons
my %top_harvest; # top harvest ladder

my %members; # clan members

# resulver
my %ip;        # ip->name(s)
my %ip_rev;    # name->ip(s)
my %guid;      # guid->name(s)
my %guid_rev;  # name->guid(s)
my %same_ip;   # name->name(s) from the same ip
my %same_guid; # name->name(s) from the same guid

#-----------------------------------------------------------------------------

my %snipe_head;
$snipe_head{ 'NAME' }         = "<1>Player";
$snipe_head{ 'ALL' }          = "<2>[>5%]ALL";
$snipe_head{ 'GARAND_SCOPE' } = "<3>[>5%]GAR";
$snipe_head{ 'K43_SCOPE' }    = "<4>[>5%]K43";
$snipe_head{ 'FG42SCOPE' }    = "<5>[>5%]FG42";

#-----------------------------------------------------------------------------

$T = time() if $PROFILE;

setpriority( 0, 0, 19 ) if ! $ENV{ 'ETS_FULL_SPEED' } and $^O ne "MSWin32";

read_log_file();

#-----------------------------------------------------------------------------

if( $log_con_file )
  {
  my $pbc = 0;
  my $cnt = 0;
  print STDERR "processing console log file: $log_con_file\n";
  open my $i, $log_con_file;
  while(<$i>)
    {
    next if /^[\s\r\n]*$/o;
    s/\^.//g;
    
    next unless /^Userinfo: \\(.+)/;
    my %ui = split /\\/, $1;
    
    $cnt++;
    $pbc++ if $ui{'cl_punkbuster'};

    my $n = $ui{'name'};
    my $g = $ui{'cl_guid'};
    my $i = $ui{'ip'};
    
    $i =~ s/:\d+$//;
    
    $ip{ $i }{ $n }++;
    $ip_rev{ $n }{ $i }++;
    
    $guid{ $g }{ $n }++;
    $guid_rev{ $n }{ $g }++;
    }
  close $i;  

  while( my ( $n, $hr ) = each %ip_rev )
    {
    for my $i ( keys %$hr )
      {
      for my $nn ( keys %{ $ip{ $i } } )
        {
        $same_ip{ $n }{ $nn }++;
        }
      }
    }
  
  while( my ( $n, $hr ) = each %guid_rev )
    {
    for my $g ( keys %$hr )
      {
      next if $g eq 'unknown';
      for my $nn ( keys %{ $guid{ $g } } )
        {
        $same_guid{ $n }{ $nn }++;
        }
      }
    }
  
  print STDERR sprintf "%.2f%% of players are using PB\n", ( 100.0 * $pbc ) / $cnt;
  }
  
print STDERR "profile: reading console log: ".(time()-$T)." seconds\n" and $T = time() if $PROFILE;

#-----------------------------------------------------------------------------

print STDERR "processing data...\n";

$nam{ '*WORLD*' }++;
# remove fake players
for my $k ( keys %nam )
  {
  next if $k eq '*WORLD*';
  next if $k =~ /^!/;
  next if $k =~ /^MAP:/;
  delete $nam{ $k } if $k eq 'unknown';
  # never reached XP above 100
  delete $nam{ $k } unless $player{ $k }{ 'TOP_SCORE' } > $SCORE_FLOOR;
  }

$nam_color{ 'AXIS' } = "^1AXIS";
$nam_color{ 'ALLIES' } = "^4ALLIES";
$nam_color{ 'SPECTATORS' } = "^2SPECTATORS";

# fix names links
for my $k ( keys %nam )
  {
  my $k2 = $k;
  $k2 =~ s/[^!a-z0-9\-\_]/_/gi;
  $k2 =~ s/[!]/./gi;
  $href{ $k } = $k2;
  
  my $c = get_clan( $k );
  $members{ $c }{ $k }++;
  }

print STDERR "profile: processing data: ".(time()-$T)." seconds\n" and $T = time() if $PROFILE;

#-----------------------------------------------------------------------------

if( $log_pb_file )
  {
  my $pbc = 0;
  my $cnt = 0;
  print STDERR "processing PB log file: $log_pb_file\n";
  
  my @data;
  
  my %seen;
  
  open my $i, $log_pb_file;
  while(<$i>)
    {
    s/[\r\n]+$//o;
    # [10.21.2003 18:14:40] VIOLATION (WALLHACK) #60184: ^$WARIOR (slot #2) Violation (WALLHACK) #60184 [9d92d8d22b8fc763555a4f0683b7af1c(-) 217.10.250.23:27960]
    /\[([^\]]+)\]\s*VIOLATION\s*\(([^:]+)\)\s*#(\d+):\s*(.+?)\s*\(slot #\d+\).+?([\d\.\:]+)\]$/io;
    my $time = $1;
    my $viol = $2;
    my $vid  = $3;
    my $name = $4;
    my $addr = $5;
    
    $name =~ s/\^.//g;
    
    next unless $vid >= 50_000; # PB docs says violations over 50k are cheats
    
    next if $seen{ "$viol|$name" }++;
    
    # print STDERR "$time | $viol | $name | $addr\n";
    my $cname = color_name( $name );
    unshift @data, [ '', "<a href=!p_$href{$name}.html>$cname</a>", $viol, $vid, $name, $time, $addr ];
    pop @data if @data > 50;
    }
  close( $i );

  for( @data )
    {
    my $k = $_->[2] . '|' . $_->[4];
    $_->[4] = $seen{ $k };
    }
  
  unshift @data, [ 'HEAD', "[30%]Player", "Violation", "VID", "Count", "Time", "IP" ];
  html_save( 'punks', "<h3>hall of shame (busted punks)</h3>", html_table( \@data, SORT => 4 ) );
  }
else
  {
  html_save( 'punks', "<h3>hall of shame (busted punks)</h3>", 'no information available' );
  }
print STDERR "profile: reading PB log: ".(time()-$T)." seconds\n" and $T = time() if $PROFILE;

#-----------------------------------------------------------------------------

my %kpm_temp_game;
while( my ( $k, $v ) = each %kpm_temp )
  {
  my ( $max_k, $max_v ) = hash_max( $v );
  # $k = sprintf "%s  %s", scalar localtime( $max_k ), color_name( $k );
  $kpm{ $k } = $max_v;
  my $ti = $kpm_game{ $max_k };
  $kpm_temp_game{ $k } = $ti;
  }
%kpm_game = %kpm_temp_game;

print STDERR "done\n";

print STDERR "profile: calculating KPM: ".(time()-$T)." seconds\n" and $T = time() if $PROFILE;

#-----------------------------------------------------------------------------

exit if $PROFILE;

print STDERR "writing statistics...\n";

my @games_text;

push @games_text, [ 'HEAD', 
                    '[>5%]#', 
                    '[15%]Map',
                    '[25%]Finish Time', 
                    '[>5%]IXP', 
                    '[>5%]AXP', 
                    '[5%]TLH',
                    '[40%]Players' ];
@games = reverse @games;
my $gc = 0; # game counter
for my $gr ( @games )
  {
  my $ti = $gr->{ 'TIME' };
  my $ma = $gr->{ 'MAP' };
  my $tm = localtime( $ti );
  $tm =~ s/:\d\d / /; # remove seconds
  my $pl = sprintf "%02d", $gr->{ 'PLAYERS' };
  next unless $pl >= 4;
  my $ww = $pl * 8;

  my $snipe = $gr->{ 'SNIPE' };
  $snipe->{ 'HEAD' } = \%snipe_head;
  for my $p ( keys %$snipe )
    {
    my $pc = color_name( $p );
    $snipe->{ $p }{ 'NAME' } = "<a class=t href=!p_$href{$p}.html>$pc</a>";
    }
  
  # now fix some links and info
  $games{ $ti } = $gr;
  my $des = $gr->{ 'DES' } = "$tm, $pl [T:pl.] [T:on] $ma";
  
  push @games_text, [ '', 
                      '(ROW)', 
                      $ma, 
                      "<a href=!game_$ti.html>$tm</a>&nbsp;(<a href=!game_$ti\_trans.html>T</a>)",
                      $gr->{ 'AXIS_XP' },
                      $gr->{ 'ALLIES_XP' },
                      $gr->{ 'TIMELIMIT' } ? 'Yes' : 'No',
                      "$pl <img src=#$theme/img/wp.gif height=10 width=$ww>"
                    ];
  html_save( "game_$ti", "<h2>[T:game at] $des</h2>",
                         mod_hash_stats_2d( 'Captures/Dynamites/Repairs', $gr->{ 'ITEMS' } ),
                         single_game_stats( $gr, TYPE => 'SKILLS', SIDE => 'ALLIES' ),
                         single_game_stats( $gr, TYPE => 'SKILLS', SIDE => 'AXIS' ),
                         single_game_stats( $gr, TYPE => 'SHOTS', SIDE => 'ALLIES' ),
                         single_game_stats( $gr, TYPE => 'SHOTS', SIDE => 'AXIS' ),
                         single_game_stats( $gr, TYPE => 'DAMAGE', SIDE => 'ALLIES' ),
                         single_game_stats( $gr, TYPE => 'DAMAGE', SIDE => 'AXIS' ),
                         mod_hash_stats_2d( "[T:sniper kills]", $snipe, SORT => 2, FLOOR => 16 ),
                         single_game_stats( $gr, TYPE => 'DETAILS' ),
                         "<p><a href=!game_$ti\_trans.html>[T:Transcription]</a>" );
  html_save( "game_$ti\_trans", "<h2>[T:game at] $tm [T:on] $ma [T:map], $pl [T:players]</h2>",
                         single_game_transcription( $gr, TYPE => 'TRANS' ) );
  $gc++;
  }

html_save( 'index', $html_motd );
html_save( 'games', $html_motd, "<p><h2>[T:Last games: (approx. last 24 hours)]</h2><p>", 
                                html_table( [ grep /\S/, @games_text[0..50] ] ), 
                                "<p><a href=!games_all.html>[T:show all] $gc [T:games (long!)]</a>" );
html_save( 'games_all',         $html_motd . "<p><h2>[T:All] $gc [T:games]</h2><p>" . html_table( \@games_text, FLOOR => 1_000_000 ) );

html_save( "players", mod_players( CLAN => 0 ) );
html_save( "clans",   mod_players( CLAN => 1 ) );

html_save( 'p_sides' , make_columns(
                       mod_hash_stats( "AXIS [T:(totals per clan)]", $sides{ 'AXIS' }, CLAN => 1, FLOOR => 16 ),
                       mod_hash_stats( "ALLIES [T:(totals per clan)]", $sides{ 'ALLIES' }, CLAN => 1, FLOOR => 16 ) ),
                       make_columns(
                       mod_hash_stats( "AXIS", $sides{ 'AXIS' }, FLOOR => 16 ),
                       mod_hash_stats( "ALLIES", $sides{ 'ALLIES' }, FLOOR => 16 ) ) );

for my $k ( keys %nam )
  {
  my $text;
  my $clan = 0;
  $clan = 1 if $k =~ /^!/;
  $text .= mod_clan_members( "[T:clan members of] " . color_name( $k ), $k ) if $clan;
  
  $text .= make_columns( mod_hash_stats( "[T:sides played by] " . color_name( $k ), $side{ $k } ),
                         mod_hash_stats( "[T:classes played by] " . color_name( $k ), $class{ $k } ) );
  
  $text .= mod_top_scores_stats( "[T:top scores for] " . color_name( $k ), $k );
  $text .= mod_player_weapons_stats( "[T:weapons' totals for] " . color_name( $k ), $k );
  $text .= mod_kill_killed( "[T:public relations for] " . color_name( $k ), $k );
  $text .= mod_games_for_player( "[T:last games in which participated] " . color_name( $k ), $k );
  
  $text .= "<p><a href=!p_r_$href{$k}.html>([T:resolver])</a>" if $log_con_file;

  html_save( 'p_' . $href{$k}, $text );
  
  next unless $log_con_file;
  
  $text = undef;
  
  $text .= mod_hash_stats( "[T:aliases for player] " . color_name( $k ), $same_guid{ $k } );
  $text .= mod_hash_stats( "[T:players from the same ip address(es)] " . color_name( $k ), $same_ip{ $k } );
  
  html_save( 'p_r_' . $href{$k}, $text );
  }

html_save( 'noise',    make_columns(
                       mod_hash_stats( "[T:noise in words by clan]", \%noise, CLAN => 1, FLOOR => 16 ),
                       mod_hash_stats( "[T:noise in words by player]", \%noise, FLOOR => 16 ),
                       ) );

for my $p ( keys %top_snipe )
  {
  my $pc = color_name( $p );
  $top_snipe{ $p }{ 'NAME' } = "<a class=t href=!p_$href{$p}.html>$pc</a>";
  }

$top_snipe{ 'HEAD' } = \%snipe_head;
html_save( 'covert',   mod_hash_stats_2d( "[T:top sniper kills in single game]", \%top_snipe, SORT => 2, FLOOR => 16 ),
                       make_columns(
                       mod_hash_stats_2d( "[T:GARAND]", \%top_snipe, SORT => 3, FLOOR =>  8 ),
                       mod_hash_stats_2d( "[T:K43]", \%top_snipe, SORT => 4, FLOOR =>  8 ),
                       mod_hash_stats_2d( "[T:FG42]", \%top_snipe, SORT => 5, FLOOR =>  8 ),
                       ),
                       make_columns(
                       mod_hash_stats( "[T:uniforms taken]", \%unif_taken, FLOOR => 16 ),
                       mod_hash_stats( "[T:uniforms lost]", \%unif_lost, FLOOR => 16 ),
                       ) );

html_save( 'kpm',      make_columns(
                       mod_kpm_stats( "[T:top kills per minute by clan]", CLAN => 1, FLOOR => 16 ),
                       mod_kpm_stats( "[T:top kills per minute by player]", FLOOR => 16 ),
                       ) );

html_save( 'waste',    make_columns(
                       mod_hash_stats( "[T:time wasted by clan]",   \%time_waste, TIMEFMT => 1, CLAN => 1, HEAD => "#,Player,Time,Percentage", FLOOR => 16 ),
                       mod_hash_stats( "[T:time wasted by player]", \%time_waste, TIMEFMT => 1,  HEAD => "#,Clan,Time,Percentage", FLOOR => 16 ),
                       ) );

html_save( 'classide', make_columns(
                       mod_hash_stats( "[T:sides played]", \%global_side ),
                       mod_hash_stats( "[T:classes played]", \%global_class ) ) );

html_save( 'top_td',   make_columns(
                       mod_hash_stats( "[T:'no friends' ladder by clan]",   \%top_td, NOP => 1, FLOOR => 16, CLAN => 1 ),
                       mod_hash_stats( "[T:'no friends' ladder by player]", \%top_td, NOP => 1, FLOOR => 16 ) ) );

html_save( 'harvest',  "<h2>[T:top figures for the harvest ladder]</h2>",
                       make_columns(
                       mod_hash_stats( "[T:health packs dispensed by player]", $top_harvest{ 'HEALTH_PACK/DISPENSED' }, FLOOR => 16 ),
                       mod_hash_stats( "[T:health packs collected by player]", $top_harvest{ 'HEALTH_PACK/COLLECTED' }, FLOOR => 16 ),
                       ),
                       make_columns(
                       mod_hash_stats( "[T:health packs dispensed by clan]", $top_harvest{ 'HEALTH_PACK/DISPENSED' }, FLOOR => 16, CLAN => 1 ),
                       mod_hash_stats( "[T:health packs collected by clan]", $top_harvest{ 'HEALTH_PACK/COLLECTED' }, FLOOR => 16, CLAN => 1 ),
                       ),
                       make_columns(
                       mod_hash_stats( "[T:ammo packs dispensed] by player", $top_harvest{ 'AMMO_PACK/DISPENSED' }, FLOOR => 16 ),
                       mod_hash_stats( "[T:ammo packs collected] by player", $top_harvest{ 'AMMO_PACK/COLLECTED' }, FLOOR => 16 ),
                       ),
                       make_columns(
                       mod_hash_stats( "[T:ammo packs dispensed] by clan", $top_harvest{ 'AMMO_PACK/DISPENSED' }, FLOOR => 16, CLAN => 1 ),
                       mod_hash_stats( "[T:ammo packs collected] by clan", $top_harvest{ 'AMMO_PACK/COLLECTED' }, FLOOR => 16, CLAN => 1 ),
                       ), );

my $skills_totals = mod_skills_totals( "[T:score totals]" );
for my $c ( '', 4..11 )
  {
  html_save( "tops$c",   "[T:choose]:
                            <a href=!tops.html>[T:XP]</a>
                          | <a href=!tops4.html>[T:battle sense]</a>
                          | <a href=!tops5.html>[T:engineer]</a>
                          | <a href=!tops6.html>[T:medic]</a>
                          | <a href=!tops7.html>[T:field ops]</a>
                          | <a href=!tops8.html>[T:light weapons]</a>
                          | <a href=!tops9.html>[T:heavy weapons (soldiers)]</a>
                          | <a href=!tops10.html>[T:covert ops]</a>
                          | <a href=!tops11.html>[T:objectives]</a>",
                        mod_tops( "[T:top scores in single game]", SORT => $c || 3 ),
                        $skills_totals,
                        );
  }

for my $c ( '', 3..9 )
  {
  html_save( "wep$c",   "[T:sort by]:
                            <a href=!wep3.html>[T:Shots]</a>
                          | <a href=!wep4.html>[T:Hits]</a>
                          | <a href=!wep5.html>[T:Head]</a>
                          | <a href=!wep6.html>[T:Kills]</a>
                          | <a href=!wep7.html>[T:Acc.%]</a>
                          | <a href=!wep8.html>[T:HeadAcc.%]</a>",
                        mod_weapons_totals( "[T:weapons totals]", SORT => $c || 6 ) );
  }

for my $w ( keys %WEAPON )
  {
  my $ww = lc $WEAPON{ $w };
  for my $c ( '', 3..9 )
    {
    html_save( "w_$ww$c", "[T:sort by]: 
                            <a href=!w_${ww}3.html>[T:Shots]</a>
                          | <a href=!w_${ww}4.html>[T:Hits]</a>
                          | <a href=!w_${ww}5.html>[T:Head]</a>
                          | <a href=!w_${ww}6.html>[T:Kills]</a>
                          | <a href=!w_${ww}7.html>[T:Deaths]</a>
                          | <a href=!w_${ww}8.html>[T:Acc.%]</a>
                          | <a href=!w_${ww}9.html>[T:HeadAcc.%]</a>",
                          # mod_weapons_usage( "[T:usage of weapon] $WEAPON{$w} [T:by clan]", $w, CLAN => 1, SORT => $c || 6),
                          mod_weapons_usage( "[T:top figures for weapon] $WEAPON{$w} [T:by player]", $w, CLAN => 0, SORT => $c || 6 ) );
    }                    
  }                  

html_save( 'p_',    'no info' ),

print STDERR "profile: writing output: ".(time()-$T)." seconds\n" and $T = time() if $PROFILE;

exit;

##############################################################################

sub read_log_file
{
  my $time_now = time();
  
  my $time_base = 0;
  my $local_time = 0;
  my $last_local_time = 0;
  my $time = 0;
  
  my $last_game_time;
  my %last_server;
  my %stats;    # game statistics, temporary used to collect data
  my %items;    # items collected/processed during the game, temporary used to collect data
  my %harvest;  # harvest awards, ammo and medkits
  my %snipe;    # usage of sniper weapons
  my %referees; # hold referees' names
  my $exit = 0; # keep status wether the level is over
  
  $last_days = $last_days * 24 * 60 * 60; # convert days into seconds
  
  print STDERR "processing log file: $log_file\n";
  
  open( my $i, $log_file ) or die "$0: cannot open log file: $log_file\n";
  while(<$i>)
    {
    # print "line: $.\n" if $. % 10000 == 0;
    next if /^[\s\r\n]*$/o;
    s/<world>/*WORLD*/gio;
    s/^"?\s*(\d+):(\d+)(\.\d+)?\s*//o; # get time (^"? is for the buggy shrubet teamdamage logging)
#    use integer;
    $local_time = $1 * 60 + $2;
    if( $last_local_time > $local_time )
      {
      $time_base = $time_base + $last_local_time;
      }
    $last_local_time = $local_time;
    $time_base = $1 if /^TIME:(\d+)/o; # get time base
    $time = $time_base + $local_time;
    next if $last_days > 0 and $time_now - $last_days > $time;
#    no integer;
    $minute = $time - $time % 60;
    my $time_str = scalar localtime( $time );
    $time_str =~ s/  / 0/o;
  
  #print STDERR "$time_str ($time) [$time_base:$local_time] $_";
  
    if( /^ClientConnect: (\d+)/io )
      {
      $TC{ 'CDo' }++ if $PROFILE;
      #print STDERR "c: $1\n";
      next if exists $conn_time{ $1 };
      $conn_time{ $1 } = $time;
      next;
      }
    if( /^ClientDisconnect: (\d+)/io )
      {
      $TC{ 'CDd' }++ if $PROFILE;
      next unless exists $conn_time{ $1 };
      my $wasted = int( $time - $conn_time{ $1 } );
      delete $conn_time{ $1 };
      my $name = $client{ $1 }{ 'n' };
      #print STDERR "d: $1 ($name) $wasted\n";
      next if $name eq ''; # anonymous
      $time_waste{ $name } += $wasted;
      $time_waste{ get_clan( $name ) } += $wasted;
      next;
      }
    if( /^ShutdownGame:/o )
      {
      $TC{ 'SG' }++ if $PROFILE;
      for my $k ( keys %conn_time )
        {
        my $wasted = int( $time - $conn_time{ $k } );
        my $name = $client{ $k }{ 'n' };
        next if $name eq '';
        $time_waste{ $name } += $wasted;
        $time_waste{ get_clan( $name ) } += $wasted;
        #print STDERR "d: $k ($name) $wasted\n";
        }
      %conn_time = ();  
      # next; # nadolu ima oshte edin handler za shutdown game
      }
    if( /^ClientUserinfoChanged: (\d+) (.+)/io )
      {
      $TC{ 'CUC' }++ if $PROFILE;
      # n\xwayBG\t\2\c\4\r\0\m\0000000\s\0000000\dn\chromosome\dr\0\w\10\lw\10\sw\41\mu\0\ref\0
      my $id = $1;
      my $data = $2;
      $client{ $id } = { split /\\/, $data };
      
      my ( $name, $color_name );
      $name = $color_name = $client{ $id }{ 'n' };
      $name =~ s/\^.//go; # clear colors etc.
      
      $client{ $1 }{ 'n' } = $name = $name_map{ $name }
          if( $name_map{ $name } );
      $nam{ $name }++;
      $nam{ get_clan( $name ) }++;
      
      $nam_color{ $name } = $color_name if $name ne $color_name;
      
      $client{ $id }{ 'n' } = $name;
      $client{ $id }{ 'dn' } =~ s/\^.//go; # clear colors etc.
  
      my $dn     = $client{ $id }{ 'dn' };
      my $class  = $CLASS{ $client{ $id }{ 'c' } };
      my $side   = $SIDE{ $client{ $id }{ 't' } };
      my $weapon = $WEAPON{ $client{ $id }{ 'w' } };
      my $ref    = $client{ $id }{ 'ref' };
      
      my $info;
      $info .= "$time_str ^2$name has referee status" if $ref and ! $referees{ $id }++;
      push @trans, $info if $info;
      
      if( $dn )
        {
        $unif_taken{ $name }++;
        $unif_lost{ $dn }++;
        };

      if( $class_last{ $name } ne $class )
        {
        $class{ $name }{ $class }++;
        $class{ get_clan( $name ) }{ $class }++;
        $class_last{ $name } = $class;
        $global_class{ $class }++;
        $classes{ $class }{ $name }++;
        $classes{ $class }{ get_clan( $name ) }++;
        }
      if( $side_last{ $name } ne $side )
        {
        $side{ $name }{ $side }++;
        $side{ get_clan( $name ) }{ $side }++;
        $side_last{ $name } = $side;
        $global_side{ $side }++;
        $sides{ $side }{ $name }++;
        $sides{ $side }{ get_clan( $name ) }++;
        }

      next;
      }
    s/\^.//g; # clear colors etc. this is after ClientUserinfoChanged to preserve original names
    if( /^Item: (\d+) team_CTF_(red|blue)flag/o )
      {
      my $n = $client{ $1 }{ 'n' };
      my $f = uc $2;
      
      $items{ $n }{ "CAPTURES/FLAGS/$f" }++;
      $items{ $n }{ "TOTAL" }++;
      next;
      }
    if( /^Dynamite_Plant: +(\d+)/o )
      {
      my $n = $client{ $1 }{ 'n' };
      $items{ $n }{ 'DYNA/PLANT' }++;
      $items{ $n }{ "TOTAL" }++;
      next;
      }
    if( /^Dynamite_Diffuse: +(\d+)/o )
      {
      my $n = $client{ $1 }{ 'n' };
      $items{ $n }{ 'DYNA/DIFF' }++;
      $items{ $n }{ "TOTAL" }++;
      next;
      }
    if( /^Repair: (\d+)/o )
      {
      my $n = $client{ $1 }{ 'n' };
      $items{ $n }{ 'REPAIR' }++;
      $items{ $n }{ "TOTAL" }++;
      next;
      }
    if( /^([a-z]+_Pack): (\d+) (\d+)/io )
      {
      my $f = $client{ $2 }{ 'n' }; # from 
      my $t = $client{ $3 }{ 'n' }; # to
      $harvest{ uc "$1/DISPENSED" }{ $f }++;
      $harvest{ uc "$1/COLLECTED" }{ $t }++;
      $harvest{ uc "$1/DISPENSED" }{ get_clan( $f ) }++;
      $harvest{ uc "$1/COLLECTED" }{ get_clan( $t ) }++;
      }
      
    if( $exit and /WeaponStats:\s*(.+)$/o )
      {
      $TC{ 'WS' }++ if $PROFILE;
      my $ws = $1;
      my $data = parse_weaponstats( $ws, $last_game_time );
      $stats{ $data->{ 'PLAYER' } } = $data if $data;
      #my $map = uc "MAP:" . $last_server{ 'mapname' };
      #$nam{ $map }++;
      #parse_weaponstats( $ws, $last_game_time, PLAYER => $map );
      next;
      }
    if( /^InitGame:\s*(.*)$/o )
      {
      $TC{ 'IG' }++ if $PROFILE;
      $exit = 0;
      $last_game_time = $time;
      %last_server = split /\\/, "\\$1";
      next;
      }
    if( /Exit:/o )
      {
      $TC{ 'EX' }++ if $PROFILE;
      $last_server{ 'TIMELIMIT' } = /Timelimit/ ? 1 : 0;
      $exit = 1;
      }
    if( $exit and /^ShutdownGame/o )
      {
      $TC{ 'SG' }++ if $PROFILE;
      
      # format items
      $items{ 'HEAD' }{ 'NAME' } = '<1>[100%]Player';
      $items{ 'HEAD' }{ 'CAPTURES/FLAGS/RED' } = "<2>[>]Red Flag";
      $items{ 'HEAD' }{ 'CAPTURES/FLAGS/BLUE' } = "<3>[>]Blue Flag";
      $items{ 'HEAD' }{ 'DYNA/PLANT' } = "<4>[>]Dyna Plant";
      $items{ 'HEAD' }{ 'DYNA/DIFF' } = "<5>[>]Dyna Diff";
      $items{ 'HEAD' }{ 'REPAIR' } = "<6>[>]Repairs";
      $items{ 'HEAD' }{ 'TOTAL' } = "<7>[>]Total";
      
      $items{ 'FOOT' }{ 'NAME' } = 'Totals:';
      $items{ 'FOOT' }{ 'CAPTURES/FLAGS/RED' } = '(SUM)/i';
      $items{ 'FOOT' }{ 'CAPTURES/FLAGS/BLUE' } = '(SUM)/i';
      $items{ 'FOOT' }{ 'DYNA/PLANT' } = '(SUM)/i';
      $items{ 'FOOT' }{ 'DYNA/DIFF' } = '(SUM)/i';
      $items{ 'FOOT' }{ 'REPAIR' } = '(SUM)/i';
      $items{ 'FOOT' }{ 'TOTAL' } = '(SUM)/i';
  
      $items{ $_ }{ 'NAME' } = color_name( $_ ) || '*unknown*' for grep ! /^(HEAD|FOOT|COLS)$/, keys %items;
      
      my @players = keys %stats;
      my $pc = scalar @players;
      
      if( $pc >= 4 )
        {
        push @games, { 'STATS' => { %stats },
                       'ITEMS' => { %items },
                       'TIME' => $last_game_time, 
                       'PLAYERS' => $pc,
                       'MAP' => $last_server{ 'mapname' },
                       'AXIS_XP' => $last_server{ 'AXIS_XP' },
                       'ALLIES_XP' => $last_server{ 'ALLIES_XP' },
                       'TIMELIMIT' => $last_server{ 'TIMELIMIT' },
                       'SNIPE' => { %snipe },
                       'TRANS' => [ @trans ],
                        };
        
        for my $p ( grep ! /^(HEAD|FOOT|COLS)$/, keys %items )
          {
          my $v;
          $v  = $items{ $p }{ 'TOTAL' };
          $v .= '.' . sprintf "%04d", $stats{ $p }{ 'SCORE' };
          # top objectives
          if( $v > $player{ $p }{ 'TOP_OBJ' } )
            {
            $player{ $p }{ 'TOP_OBJ' } = $v;
            $player{ $p }{ 'TOP_OBJ_GAME' } = $last_game_time;
            }
          }
        
        for my $p ( keys %snipe )
          {
          for my $w ( keys %{ $snipe{ $p } } )
            {
            $top_snipe{ $p }{ $w } = $snipe{ $p }{ $w }
                if $snipe{ $p }{ $w } > $top_snipe{ $p }{ $w };
            }
          }
          
        for my $p ( keys %harvest )
          {
          for my $n ( keys %{ $harvest{ $p } } )
            {
            $top_harvest{ $p }{ $n } = $harvest{ $p }{ $n }
                if $harvest{ $p }{ $n } > $top_harvest{ $p }{ $n };
            }
          }
          
        }                
      @trans = ();
      %stats = ();
      %snipe = ();
      %items = ();
      %harvest  = ();
      %referees = ();
      next;
      }  
    if( /red:(\d+) +blue:(\d+)/ )
      {
      $last_server{ 'AXIS_XP' } = $1;
      $last_server{ 'ALLIES_XP' } = $2;
      next;
      }
    if( /Kill: (\d+) (\d+) (\d+): (.+) killed (.+) by MOD_([A-Z_0-9]+)/io )
      {
      $TC{ 'KI' }++ if $PROFILE;
      my $k = $name_map{ $4 } || $4;
      my $d = $name_map{ $5 } || $5;
      my $w = $6;
      
      $kpm_temp{ $k }{ $minute }++;
      $kpm_temp{ get_clan( $k ) }{ $minute }++;
      $kpm_game{ $minute } = $last_game_time;
      
      # kills
      $kil{ $k }{ $d }++;
      $kil{ get_clan( $k ) }{ get_clan( $d ) }++;
      
      # deaths
      $ded{ $d }{ $k }++;
      $ded{ get_clan( $d ) }{ get_clan( $k ) }++;
      
      # weapon usage
      if( $w =~ /SCOPE$/ )
        {
        $snipe{ $k }{ $w }++;
        $snipe{ $k }{ 'ALL' }++;
        }
=pod    
      # get names...
      $nam{ $k }++;
      $nam{ $d }++;
      $nam{ get_clan( $k ) }++;
      $nam{ get_clan( $d ) }++;
=cut
  
      next;
      }
    if( /^say(team)?: (.+?): (.+)/io )
      {
      my $clr = $1 ? '^3' : '^5';
      
      $TC{ 'SA' }++ if $PROFILE;
      my $p = $2;
      my $cnt = $3 =~ /\S+/g;
      $noise{ $p } += $cnt;
      $noise{ get_clan( $p ) } += $cnt;
      
      push @trans, "$time_str $clr$_";
      next;
      }
    #if( /^score: +(\d+) +ping: +(\d+) +client: (\d+) (.+)/ )
    #  {
    #  next;
    #  }
    #if( /?? end/start of game ??/ )
    #  {
    #  %class_last = ();
    #  %side_last = ();
    #  next;
    #  }  
    }
  close( $i );
  
  if( $PROFILE )
    {
    use Data::Dumper;
    print STDERR "profile counters:\n" . Dumper( \%TC ) . "\n";
    }
  
  print STDERR "done (end virtual time is $time)\n";
  print STDERR "profile: reading main log: ".(time()-$T)." seconds\n" and $T = time() if $PROFILE;
}


##############################################################################

sub make_columns
{
  my $text;
  
  my $c = @_;
  my $p = $c > 0 ? int( 100 / @_ ) : 100;
  
  $text .= "<table cellspacing=5 width=100%><tr><td width=p% valign=top>";
  $text .= join "</td><td width=$p% valign=top>", @_;
  $text .= "</td></tr></table>";
  
  return $text;
}

sub mod_tops
{
  my $title = shift;
  my %opt = @_;
  my $clan = $opt{ 'CLAN' };
  
  my @data;

  my @p = keys %nam;
  @p = grep /^!/, @p if $clan;
  @p = grep ! /^!/, @p if ! $clan;
  
  push @data, [ 'HEAD', "[>5%]#", "[30%]Player/Clan", 
                        "[>7%]XP",
                        map( { '[>7%]' . $SKILL_IMG{ $_ } } ( 0..6 ) ),
                        "[>7%]OBJ", 
                         ];
  
  my $r = 0;
  for my $p ( @p )
    {
    my $pc = color_name( $p );
    
    my $og = $player{ $p }{ 'TOP_OBJ_GAME' };
    my $ob = $player{ $p }{ 'TOP_OBJ' };
    
    my $ti = $player{ $p }{ 'TOP_SCORE_GAME' };
    my $xp = $player{ $p }{ 'TOP_SCORE' };
    
    my @sk = map( { $player{ $p }{ 'TOP_SKILL' }{ $_ } } ( 0..6 ) );
    my @sg = map( { $player{ $p }{ 'TOP_SKILL_GAME' }{ $_ } } ( 0..6 ) );
    
    push @data, [ '', "(ROW)", "<a class=t href=!p_$href{$p}.html>$pc</a>", 
                               "$xp<a href=game_$ti\.html>*</a>",
                               map( { $sk[$_] > 0 ? "$sk[$_]<a href=game_$sg[$_].html>*</a>" : undef } ( 0..6 ) ),
                               $ob > 0 ? "$ob<a href=game_$og\.html>*</a>" : undef,
                               ];
    }
  
  return "<h2>$title</h2>" . html_table( \@data, SORT => $opt{ SORT }, FLOOR => $TOPS_FLOOR );
}

##############################################################################

sub mod_clan_members
{
  my $title = shift;
  my %opt = @_;
  my $clan = shift;
  
  my @data;

  my @pl = keys %{ $members{ $clan } };
  
  push @data, [ 'HEAD', "[20%]Player", "[20%]Player", "[20%]Player", "[20%]Player", "[20%]Player" ];

  while( @pl )
    {
    my @r;
    for( 1..5 )
      {
      my $p = shift @pl;
      my $pc = color_name( $p );
      push @r, "<a class=t href=!p_$href{$p}.html>$pc</a>";
      }
    push @data, [ '', @r ];
    }
  
  return "<h2>$title</h2>" . html_table( \@data );
}

##############################################################################

sub mod_players
{
  my %opt = @_;
  my $clan = $opt{ 'CLAN' };

  my @data;

  my $text;
  
  #$text .= "<p>";
  $text .= $clan ? "<h2>Clans</h2>" : "<h2>Players</h2>";

  my @p = keys %nam;
  my @row;
  my $c;
  
  for my $p ( sort @p )
    {
    next if $clan and $p !~ /^!/;
    next if !$clan and $p =~ /^!/;

    if( $c++ % 5 == 0 )
      {
      push @data, [ '', @row ];
      @row = ();
      }
 
    my $c = get_clan( $p );

    my $target1 = $href{ $c };
    my $target2 = $href{ $p };

    $p = color_name( $p );
    $c = color_name( $c );

    if( $c eq '!' )
      {
      push @row, "[20%]&nbsp <a href=!p_$target2.html>$p</a>";
      }
    else
      {  
      push @row, "[20%]<a href=!p_$target1.html>&#169</a> <a href=!p_$target2.html>$p</a>";
      }
    }
  push( @row, '[20%]&nbsp;') for ( 0 .. 5 - @row - 1 );
  push @data, [ '', @row ] if @row;
  $text .= html_table( \@data, FLOOR => 1_000_000 );
  $text .= "\n\n\n\n";
  return $text;
}

##############################################################################

sub single_game_stats
{
  my $gr = shift; # game hash reference
  my %opt = @_;
  
  my $side = $opt{ 'SIDE' };
  my $type = $opt{ 'TYPE' };
  
  my $game = $gr->{ 'STATS' };
  
  my $text;
  
  #$text .= Dumper( $gr );
  #$text = "<pre>$text</pre>";
  
  my @data_w;
  my @data_d;
  my @data_s;

  push @data_w, [ 'HEAD', '[>1%]#', '[30%]Player', '[1%]Side', '[>i]Shots', '[>i]Hits', '[>i]Head', '[>i]Kills', '[>i]Deaths', '[>p]Acc.%', '[>p]HeadAcc.%', ];
  push @data_d, [ 'HEAD', '[>1%]#', '[30%]Player', '[1%]Side', '[>i]Damage Given', '[>i]Damage Received', '[>i]Team Damage', ];
  push @data_s, [ 'HEAD', '[>1%]#', '[30%]Player', '[1%]Side', map( { '[>7%i]' . $SKILL_IMG{ $_ } } ( 0..6 ) ), '[>12%i]SCORE' ];

  my @p = keys %$game;
  @p = grep { $game->{ $_ }{ 'SIDE' } eq 'AXIS' } @p if $side eq 'AXIS';
  @p = grep { $game->{ $_ }{ 'SIDE' } eq 'ALLIES' } @p if $side eq 'ALLIES';
  #@p = sort { $game->{ $a }{ 'SIDE' } cmp $game->{ $b }{ 'SIDE' } } @p;
  @p = sort { $game->{ $b }{ 'SCORE' } <=> $game->{ $a }{ 'SCORE' } } @p;
  
  for my $p ( @p )
    {
    my $v = $game->{ $p };
    my $wt = $v->{ 'WEAPONS_TOTALS' };
    my $pc = color_name( $p );
    my $sc = color_name( $v->{ 'SIDE' } );
    push @data_w, [ '', '(ROW)',
                "[<]<a class=t href=!p_$href{$p}.html>$pc</a>", 
                "<a class=t href=!p_sides.html>$sc</a>",
                $wt->{ 'S' },
                $wt->{ 'H' },
                $wt->{ 'E' },
                $wt->{ 'K' },
                $wt->{ 'D' },
                $wt->{ 'HA' },
                $wt->{ 'EA' },
                ];

    if( $type eq 'DETAILS' )
      {
      for my $w ( sort keys %{ $v->{ 'WEAPONS' } } )
        {
        $wt = $v->{ 'WEAPONS' }{ $w };
        push @data_w, [ 'COMT', '(ROW)',
                    '[>]' . $WEAPON{ $w },
                    "<a class=t href=!p_sides.html>$sc</a>",
                    $wt->{ 'S' },
                    $wt->{ 'H' },
                    $wt->{ 'E' },
                    $wt->{ 'K' },
                    $wt->{ 'D' },
                    $wt->{ 'HA' },
                    $wt->{ 'EA' },
                    ];
        }
      }          
    push @data_d, [ '', '(ROW)',
                "<a class=t href=!p_$href{$p}.html>$pc</a>", 
                color_name( $v->{ 'SIDE' } ),
                $v->{ 'DAMAGE' }{ 'SENT' },
                $v->{ 'DAMAGE' }{ 'RECV' },
                $v->{ 'DAMAGE' }{ 'TEAM' },
                ];
    my @ss;
    push @ss, $v->{ 'SKILL' }{ $_ } for ( 0..6 );
    push @data_s, [ '', '(ROW)',
                "<a class=t href=!p_$href{$p}.html>$pc</a>", 
                color_name( $v->{ 'SIDE' } ),
                @ss,
                $v->{ 'SCORE' }
                ];
    }

  push @data_w, [ 'FOOT', '', '', 'Total:', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(AVE)/p', '(AVE)/p', ];
  push @data_d, [ 'FOOT', '', '', 'Total:', '(SUM)/i', '(SUM)/i', '(SUM)/i', ];
  push @data_s, [ 'FOOT', '', '', 'Total:', map { '(SUM)/i' } ( 0 .. 6 ), '(SUM)/i' ];
  
  $side = color_name( $side );
  $text .= "<h3>$side Shots and accuracy</h3>" . html_table( \@data_w ) if $type eq 'SHOTS';
  $text .= "<h3>$side Damage</h3>" . html_table( \@data_d ) if $type eq 'DAMAGE';
  $text .= "<h3>$side Skills</h3>" . html_table( \@data_s ) if $type eq 'SKILLS';
  $text .= "<h3>Detailed Shots and accuracy</h3>" . html_table( \@data_w ) if $type eq 'DETAILS';
  
  return $text;
}

sub single_game_transcription
{
  my $gr = shift; # game hash reference
  my %opt = @_;

  my $text;

  my @data_t;
  for( @{ $gr->{ 'TRANS' } } )
    {
    push @data_t, [ '', '(ROW)', color_name( $_ ) ];
    }
  $text .= "<h3>Transcription</h3>" . html_table( \@data_t, FLOOR => 1_000_000 );
  
  return $text;
}

#----------------------------------------------------------------------------

sub mod_top_scores_stats
{
  my $title = shift;
  my $p = shift; # player
  my %opt = @_;
  
  return undef unless $player{ $p };
print STDERR "$p\n" if $p =~ /^MAP:/;  
  
  my @data_s;
  
  push @data_s, [ 'HEAD', '[>5%]#', '[<40%]Skill', '[>10%i]Top Score', 'In which game' ];
  
  for my $s ( sort { $player{ $p }{ 'SKILL' }{ $b } <=> $player{ $p }{ 'SKILL' }{ $a } } 0..6 )
    {
    my $sc = $player{ $p }{ 'TOP_SKILL' }{ $s };
    next unless $sc > 0;
    my $ti  = $player{ $p }{ 'TOP_SKILL_GAME' }{ $s };
    my $des = $games{ $ti }{ 'DES' };
    push @data_s, [ '', '(ROW)', $SKILL_IMG{ $s } . " " . $SKILL{ $s }, 
                    $sc,
                    "<a href=game_$ti\.html>$des</a>"
                    ];
    }
  my $tot;
  $tot .= $SKILL_IMG{ $_ } for ( 0..6 );
  push @data_s, [ 'FOOT', '', "$tot TOTAL XP", 
                  $player{ $p }{ 'TOP_SCORE' },
                  ''
                  ];
  
  #push @data_s, [ 'FOOT', '', '', '' ];

  return "<h2>$title</h2>" . html_table( \@data_s );
}

#----------------------------------------------------------------------------

sub mod_skills_totals
{
  my $title = shift;
  my %opt = @_;
  
  my @data_s;
  
  push @data_s, [ 'HEAD', '[>5%]#', 'Skill', '[>20%i]Score' ];
  
  for my $s ( sort { $skills{ $b } <=> $skills{ $a } } keys %skills )
    {
    push @data_s, [ '', '(ROW)', $SKILL_IMG{ $s } . " " . $SKILL{ $s }, $skills{ $s } ];
    }
  
  my $tot;
  $tot .= $SKILL_IMG{ $_ } for ( 0..6 );
  push @data_s, [ 'FOOT', '', "[<]$tot TOTAL XP", '(SUM)/i' ];

  return "<h2>$title</h2>" . html_table( \@data_s );
}

#----------------------------------------------------------------------------

sub mod_player_weapons_stats
{
  my $title = shift;
  my $p = shift; # player
  my %opt = @_;
  
  return undef unless $player{ $p };
  
  my @data_w;
  
  push @data_w, [ 'HEAD', '[>5%]#', '[40%]Weapon', '[>i]Shots', '[>i]Hits', '[>i]Head', '[>i]Kills', '[>i]Deaths', '[>]K:D', '[>p]Acc.%', '[>p]HeadAcc.%', ];
  
  for my $w ( sort { $player{ $p }{ 'TOP_WEAPONS' }{ $b }{ 'K' } <=> $player{ $p }{ 'TOP_WEAPONS' }{ $a }{ 'K' } } keys %{ $player{ $p }{ 'TOP_WEAPONS' } } )
    {
    my $wt = $player{ $p }{ 'WEAPONS_TOTALS' }{ $w };
    $wt->{ 'HA' } = $wt->{ 'S' } > 0 ? ( 100 * $wt->{ 'H' } ) / $wt->{ 'S' } : 'n/a';
    $wt->{ 'EA' } = $wt->{ 'S' } > 0 ? ( 100 * $wt->{ 'E' } ) / $wt->{ 'S' } : 'n/a';
    my $kdr = $wt->{ 'D' } > 0 ? sprintf( "%.2f", $wt->{ 'K' } / $wt->{ 'D' } ) : 'n/a'; 
    my $wn = $WEAPON{ $w };
    my $ww = lc $wn;
    push @data_w, [ '', '(ROW)',
                "<a href=!w_$ww.html>$wn</a>",
                $wt->{ 'S' },
                $wt->{ 'H' },
                $wt->{ 'E' },
                $wt->{ 'K' },
                $wt->{ 'D' },
                $kdr,
                $wt->{ 'HA' },
                $wt->{ 'EA' },
                ];
    }

  push @data_w, [ 'FOOT', '', 'Total:', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(SUM)/i', '', '(AVE)/p', '(AVE)/p', ];
  
  return "<h2>$title</h2>" . html_table( \@data_w, SORT => 6 );
}

#----------------------------------------------------------------------------

sub mod_weapons_usage
{
  my $title = shift;
  my $w = shift; # weapon
  my %opt = @_;

  my $clan = $opt{ 'CLAN' };
  
  my $ww = lc $WEAPON{ $w };
  
  my @data_w;
  
  push @data_w, [ 'HEAD', '[>5%]#', '[40%]Player/Clan', "[>i]Shots", "[>i]Hits", "[>i]Head", "[>i]Kills", "[>i]Deaths", "[>p]Acc.%", "[>p]HeadAcc.%", ];
  #for my $p ( sort { $player{ $b }{ 'WEAPONS' }{ $w }{ 'K' } <=> $player{ $a }{ 'WEAPONS' }{ $w }{ 'K' } } keys %player )
  for my $p ( keys %player )
    {
    next if $clan and $p !~ /^!/;
    next if !$clan and $p =~ /^!/;

    my $wt = $player{ $p }{ 'TOP_WEAPONS' }{ $w };
    next if $wt->{ 'S' } == 0;
    eval 
      {
      $wt->{ 'HA' } = ( 100 * $wt->{ 'H' } ) / $wt->{ 'S' };
      $wt->{ 'EA' } = ( 100 * $wt->{ 'E' } ) / $wt->{ 'S' };
      };
    my $pc = color_name( $p );
    push @data_w, [ '', '(ROW)',
                "<a class=t href=!p_$href{$p}.html>$pc</a>", 
                $wt->{ 'S' },
                $wt->{ 'H' },
                $wt->{ 'E' },
                $wt->{ 'K' },
                $wt->{ 'D' },
                $wt->{ 'HA' },
                $wt->{ 'EA' },
                ];
    }
    
  push @data_w, [ 'FOOT', '', 'Total:', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(AVE)/p', '(AVE)/p', ];
  
  return "<h2>$title</h2>" . html_table( \@data_w, SORT => $opt{ 'SORT' } );
}

#----------------------------------------------------------------------------

sub mod_weapons_totals
{
  my $title = shift;
  my %opt = @_;
  
  my @data_w;
  
  push @data_w, [ 'HEAD', '[>5%]#', '[40%]Weapon', '[>i]Shots', '[>i]Hits', '[>i]Head', '[>i]Kills', '[>p]Acc.%', '[>p]HeadAcc.%', ];
  
  for my $w ( keys %weapons )
    {
    my $wt = $weapons{ $w };
    eval 
      {
      $wt->{ 'HA' } = ( 100 * $wt->{ 'H' } ) / $wt->{ 'S' };
      $wt->{ 'EA' } = ( 100 * $wt->{ 'E' } ) / $wt->{ 'S' };
      };
    my $wn = $WEAPON{ $w };
    my $ww = lc $wn;
    push @data_w, [ '', '(ROW)',
                "<a href=!w_$ww.html>$wn</a>",
                $wt->{ 'S' },
                $wt->{ 'H' },
                $wt->{ 'E' },
                $wt->{ 'K' },
                $wt->{ 'HA' },
                $wt->{ 'EA' },
                ];
    }

  push @data_w, [ 'FOOT', '', 'Total:', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(SUM)/i', '(AVE)/p', '(AVE)/p', ];
  
  return "<h2>$title</h2>" . html_table( \@data_w, SORT => $opt{ 'SORT' } );
}

##############################################################################

my %seen_weaponstats = ();
sub parse_weaponstats
{
  my $ws = shift;
  my $game_time = shift; # to skip seen weponstats per game
  my %opt = @_;
  # WeaponStats: 
  # 2 1 3145756 0 0 0 1 0 3 14 0 0 0 0 0 0 1 0 0 0 0 1 0 3 8 1 0 1 238 472 0 65 9 5
  my @ws = split /\s+/, $ws;
  
  my %data;
  
  $data{ 'CLIENT' } = shift @ws;
  $data{ 'PLAYER' } = $client{ $data{ 'CLIENT' } }{ 'n' };
  $data{ 'SIDE'   } = $SIDE{ $client{ $data{ 'CLIENT' } }{ 't' } };
  $data{ 'CLASS'  } = $CLASS{ $client{ $data{ 'CLIENT' } }{ 'c' } };
  $data{ 'CARRY' } = shift @ws; # 0 = first round, 1 == carry, i.e. second or next round

  $data{ 'PLAYER' } = $opt{ 'PLAYER' } if $opt{ 'PLAYER' };

  my $p = $data{ 'PLAYER' };
  
  next if $seen_weaponstats{ "$game_time:$ws:$p" }++;
  
  my @wep = reverse split //, sprintf '%b', shift @ws;
  my $wc = -1; # weapon counter
  for( @wep )
    {
    $wc++;
    next unless $_;
    
    my $h = $data{ 'WEAPONS' }{ $wc }{ 'H' } = shift @ws;
    my $s = $data{ 'WEAPONS' }{ $wc }{ 'S' } = shift @ws;
            $data{ 'WEAPONS' }{ $wc }{ 'K' } = shift @ws;
            $data{ 'WEAPONS' }{ $wc }{ 'D' } = shift @ws;
    my $e = $data{ 'WEAPONS' }{ $wc }{ 'E' } = shift @ws;
    eval 
      {
      $data{ 'WEAPONS' }{ $wc }{ 'HA' } = ( 100 * $h ) / $s;
      $data{ 'WEAPONS' }{ $wc }{ 'EA' } = ( 100 * $e ) / $s;
      };
    }
  for my $w ( keys %{ $data{ 'WEAPONS' } } )
    {
    for my $a ( qw( S H K D E ) )
      {
      my $wwa = $data{ 'WEAPONS' }{ $w }{ $a };
      $data{ 'WEAPONS_TOTALS' }{ $a } += $wwa;
      $player{ $p }{ 'WEAPONS_TOTALS' }{ $w }{ $a } += $wwa;
      $weapons{ $w }{ $a } += $wwa;
      
      if( $player{ $p }{ 'TOP_WEAPONS' }{ $w }{ $a } < $wwa )
        {
        $player{ $p }{ 'TOP_WEAPONS' }{ $w }{ $a } = $wwa;
        $player{ $p }{ 'TOP_WEAPONS_GAME' }{ $w }{ $a } = $game_time;
        }
      }
    }  
  eval 
    {
    $data{ 'WEAPONS_TOTALS' }{ 'HA' } = ( 100 * $data{ 'WEAPONS_TOTALS' }{ 'H' } ) / $data{ 'WEAPONS_TOTALS' }{ 'S' };
    $data{ 'WEAPONS_TOTALS' }{ 'EA' } = ( 100 * $data{ 'WEAPONS_TOTALS' }{ 'E' } ) / $data{ 'WEAPONS_TOTALS' }{ 'S' };
    };
           $data{ 'DAMAGE' }{ 'SENT' } = shift @ws;  
           $data{ 'DAMAGE' }{ 'RECV' } = shift @ws;  
  my $td = $data{ 'DAMAGE' }{ 'TEAM' } = shift @ws;  
  
  my $cp = get_clan( $p );
  $top_td{ $p }  = $td if $td > $top_td{ $p };
  $top_td{ $cp } = $td if $td > $top_td{ $cp };
  
  my @ski = reverse split //, sprintf '%b', shift @ws;
  my $sc = -1; # skill counter
  for( @ski )
    {
    $sc++;
    next unless $_;
    my $sk = shift @ws;
    $data{ 'SKILL' }{ $sc } = $sk; # each skill for this game (xp)
    $data{ 'SCORE' } += $sk; # score for this game (xp sum)
    
    # skill totals
    $skills{ $sc } += $sk;
    
    # top skills
    if( $sk > $player{ $p }{ 'TOP_SKILL' }{ $sc } )
      {
      $player{ $p }{ 'TOP_SKILL' }{ $sc } = $sk;
      $player{ $p }{ 'TOP_SKILL_GAME' }{ $sc } = $game_time;
      }
    }
  # top scores
  if( $data{ 'SCORE' } > $player{ $data{ 'PLAYER' } }{ 'TOP_SCORE' } )
    {
    $player{ $p }{ 'TOP_SCORE' } = $data{ 'SCORE' };
    $player{ $p }{ 'TOP_SCORE_GAME' } = $game_time;
    }

  return $data{ 'SCORE' } > 0 ? \%data : undef; # ako nqma tochki, ne e igral :)
}

#----------------------------------------------------------------------------

sub mod_kpm_stats
{
  my $title = shift;
  my %opt = @_;
  my $clan = $opt{ 'CLAN' };
  
  my @data;
  
  push @data, [ 'HEAD', '[>5%]#', '[<40%]Player', '[>10%i]Top Kills', 'In which game' ];
  
  for my $p ( sort { $kpm{ $b } <=> $kpm{ $a } } keys %kpm )
    {
    next if $clan and $p !~ /^!/;
    next if !$clan and $p =~ /^!/;
    
    my $pc  = color_name( $p );
    my $k   = $kpm{ $p };
    my $ti  = $kpm_game{ $p };
    my $des = $games{ $ti }{ 'DES' };
    
    push @data, [ '', 
                  '(ROW)',
                  "[<]<a class=t href=!p_$href{$p}.html>$pc</a>", 
                  $k,
                  "<a href=game_$ti\.html>$des</a>"
                  ];
    }

  return "<h2>$title</h2>" . html_table( \@data );
}

#----------------------------------------------------------------------------

sub mod_hash_stats
{
  my $title = shift;
  my $hr    = shift;
  my %opt   = @_;
  my $text;

  my $clan = $opt{ 'CLAN' };
  my $tfmt = $opt{ 'TIMEFMT' };
  my $head = $opt{ 'HEAD' };
  
  my $sum = hash_sum( $hr );
  
  my @data;
  
  $text .= "<p>";
  $text .= "<h2>$title</h2>";
  
  push @data, [ 'OPT', "width=05% align=right", "width=40%", "width=30% align=right", "width=25% align=right" ];
  if( $head )
    {
    push @data, [ 'HEAD', split /\s*,\s*/, $head ];
    }
  else
    {  
    push @data, [ 'HEAD', "#", "Name", "Value", "Percentage" ];
    }
  
  my $row = 0;
  for my $k ( sort { $hr->{ $b } <=> $hr->{ $a } } keys %$hr )
    {
    next if $clan and $k !~ /^!/;
    next if !$clan and $k =~ /^!/;
    
    my $v = $hr->{ $k };
    my $p = $sum > 0 ? sprintf "%.2f%%", 100 * $v / $sum : 'n/a'; # '
    
    if( $href{ $k } )
      {
      $k = "<a class=t href=!p_$href{$k}.html>".color_name( $k )."</a>";
      }
    else
      {
      $k = color_name( $k );
      }  
    
    $p = 'n/a' if $opt{ 'NOP' };
      
    if ( $row < $FLOOR )
      {
      $row++;
      my $class = tr_class();
      $v = time_fmt( $v ) if $tfmt;
      push @data, [ '', '(ROW)', $k, $v, $p ];
      }        
    }
  $sum = time_fmt( $sum ) if $tfmt;
  if( $opt{ 'NOP' } )
    {
    pop( @$_ ) for ( @data );
    }
  else
    {  
    push @data, [ 'FOOT', '', 'Total:', $sum, '' ];
    }
  $text .= html_table( \@data, FLOOR => $opt{ FLOOR } );
  $text .= "\n\n\n\n";
  return $text;
}

#----------------------------------------------------------------------------

sub mod_hash_stats_2d
{
  my $title = shift;
  my $hr    = shift;
  my %opt   = @_;
  my $text;

  return '' unless $hr;
  return '' unless %$hr;

  $hr->{ 'HEAD' } or die "mod_hash_stats_2d: no HEAD info";
  my $head = { %{ $hr->{ 'HEAD' } } } || {} if $hr->{ 'HEAD' };
  my $foot = { %{ $hr->{ 'FOOT' } } } || {} if $hr->{ 'FOOT' };
  my $cols = { %{ $hr->{ 'COLS' } } } || {} if $hr->{ 'COLS' };
  my @cols = keys %$head;
  while( my ( $k, $v ) = each %$head )
    {
    next unless $v =~ s/^<(\d+)>//;
    $cols->{ $k } = $1;
    $head->{ $k } = $v;
    }
  @cols = sort { $cols->{ $a } <=> $cols->{ $b } } @cols;

  my $c = 1;
  $cols->{ $_ } = $c++ for @cols;

  my $text;
  my @data;
  my @temp;
  
  my @list = grep ! /^(HEAD|FOOT|COLS)$/, keys %$hr;
  
  return '' unless @list;
  
  @temp = ( 'HEAD' );
  push @temp, $head->{ $_ } for @cols;
  push @data, [ @temp ];
  
  for my $e ( @list )
    {
    @temp = ( '' );
    push @temp, $hr->{ $e }{ $_ } for @cols;
    push @data, [ @temp ];
    }

  if( $hr->{ 'FOOT' } )
    {
    @temp = ( 'FOOT' );
    push @temp, $foot->{ $_ } for @cols;
    push @data, [ @temp ];
    }

  $text .= "<h2>$title</h2>";
  $text .= html_table( \@data, 'SORT'  => $opt{ 'SORT' },
                               'SORTA' => $opt{ 'SORTA' },
                               'FLOOR' => $opt{ 'FLOOR' } );
  $text .= "\n\n\n\n";
  return $text;
}

#----------------------------------------------------------------------------

sub mod_kill_killed
{
  my $title = shift;
  my $p = shift; # player
  my %opt = @_;
  
  my @data;
  push @data, [ 'HEAD', "[>5%]#", "[25%]Killed:", "[>10%]Times", "[>10%]%",
                        "[>5%]#", "[25%]Killed by:", "[>10%]Times", "[>10%]%" ];

  my @killed = sort { $kil{ $p }{ $b } <=> $kil{ $p }{ $a } } keys %{ $kil{ $p } };
  my @killby = sort { $ded{ $p }{ $b } <=> $ded{ $p }{ $a } } keys %{ $ded{ $p } };
  
  my $sum_k = hash_sum( $kil{ $p } );
  my $sum_d = hash_sum( $ded{ $p } );
  
  my $c = 0;
  while(4)
    {
    my $k = shift @killed;
    my $d = shift @killby;
    
    my $kv = $kil{ $p }{ $k };
    my $dv = $ded{ $p }{ $d };
    
    my $kp = $sum_k > 0 ? ( 100 * $kv ) / $sum_k : 0;
    my $dp = $sum_d > 0 ? ( 100 * $dv ) / $sum_d : 0;
    
    my $kc = color_name( $k );
    my $dc = color_name( $d );
    
    push @data, [ '', '(ROW)', "<a class=t href=!p_$href{$k}.html>$kc</a>", $kv, "$kp/p",
                      '(ROW)', "<a class=t href=!p_$href{$d}.html>$dc</a>", $dv, "$dp/p" ];
    
    last unless @killed or @killby;
    last if ++$c == $FLOOR;
    }
    
  push @data, [ 'FOOT', '', '[>]Total:', '(SUM)', '',
                        '', '[>]Total:', '(SUM)', '' ];
  
  return "<h2>$title</h2>" . html_table( \@data, FLOOR => 16 );
}

#----------------------------------------------------------------------------

sub mod_games_for_player
{
  my $title = shift;
  my $p = shift; # player
  my %opt = @_;
  
  my @games_text;
  push @games_text, [ 'HEAD', 
                      '[>5%]#', 
                      '[15%]Map',
                      '[25%]Finish Time', 
                      '[>5%]IXP', 
                      '[>5%]AXP', 
                      '[5%]TLH',
                      '[40%]Players' ];

  for my $gr ( @games )
    {
    next unless $gr->{ 'STATS' }{ $p };
    
    my $ti = $gr->{ 'TIME' };
    my $ma = $gr->{ 'MAP' };
    my $tm = localtime( $ti );
    $tm =~ s/:\d\d / /; # remove seconds
    my $pl = sprintf "%02d", $gr->{ 'PLAYERS' };
    my $ww = $pl * 8;
    
    my $des = $gr->{ 'DES' } = "$tm, $pl [T:pl.] [T:on] $ma";
    
    push @games_text, [ '', 
                        '(ROW)', 
                        $ma, 
                        "<a href=!game_$ti.html>$tm</a>&nbsp;(<a href=!game_$ti\_trans.html>T</a>)",
                        $gr->{ 'AXIS_XP' },
                        $gr->{ 'ALLIES_XP' },
                        $gr->{ 'TIMELIMIT' } ? 'Yes' : 'No',
                        "$pl <img src=#$theme/img/wp.gif height=10 width=$ww>"
                      ];
    }
  
  return "<h2>$title</h2>" . html_table( \@games_text );
}

#----------------------------------------------------------------------------

sub html_table_cmp
{
  my $aa = shift;
  my $bb = shift;
  my $ta = shift;
  my $tb = shift;
  my $sa = shift;
  
  return -1 if $ta eq 'HEAD';
  return  1 if $tb eq 'HEAD';
  return  1 if $ta eq 'FOOT';
  return -1 if $tb eq 'FOOT';
  
  $aa = $3 if $aa =~ /^(\[([^\]]+)\])?(.+)(\/([ip]))?$/o;
  $bb = $3 if $bb =~ /^(\[([^\]]+)\])?(.+)(\/([ip]))?$/o;
  
  return $sa ? $aa cmp $bb : $bb <=> $aa;
}

sub html_table
{
  my $ar = shift;
  my %opt = @_;
  
  my $text;
  $text .= "<table class=ets width=100% cellspacing=1 cellpadding=5>\n";
  my @or; # options row
  my @ft; # format type
  my $r = -1; # row
  
  my @ar = @$ar;
  
  my $floor = $opt{ 'FLOOR' } || $FLOOR;
  
  my $oc = $opt{ 'SORT' } || $opt{ 'SORTA' }; # sort/order column
  @ar = sort { html_table_cmp( $a->[$oc], $b->[$oc], $a->[0], $b->[0], $opt{ 'SORTA' } ) } @ar if $oc > 0;
  
  for my $rr ( @ar )
    {
    $r++ if $rr->[0] eq '';
    if( $floor > 0 )
      {
      next if $rr->[0] eq 'COMT' and $r >= $floor;
      next if $rr->[0] eq ''     and $r >= $floor;
      }
    my $c = -1; # column
    my $class = tr_class( $r );
    @or = @$rr and next if $rr->[0] eq 'OPT';
    $class = 'hh' if $rr->[0] eq 'HEAD'; # header
    $class = 'ff' if $rr->[0] eq 'FOOT'; # footer
    $class = 'cc' if $rr->[0] eq 'COMT'; # comments
    $text .= "<tr class=$class>\n";
    for( @$rr )
      {
      my $d = $_;
      $c++;
      next if $c == 0;
      
      if( $d =~ s/^\[([^\]]+)\]//o )
        {
        my $p = $1; # params
        my $o; # opts
        $o .= " align=right" if $p =~ />/o;
        $o .= " width=$1"    if $p =~ /(\d+%)/o;
        $or[$c] = $o;
        $ft[$c] = lc $1 if $p =~ /([ip])$/io;
        #print STDERR "TABLE OPT: $p --> $o\n";
        #$text .= "TABLE OPT: $p --> $o<br>";
        }
      my $f;
      $f = $ft[$c] if $rr->[0] ne 'HEAD' and $rr->[0] ne 'FOOT';
      $f = lc $1 if $d =~ s/\/([ip])$//io; # [i]nt, [p]ercent
      $d = table_sum( $ar, $c ) if $d eq '(SUM)';
      $d = table_ave( $ar, $c ) if $d eq '(AVE)';
      $d = table_med( $ar, $c ) if $d eq '(MED)';
      
      $d = num_fmt( int $d ) if $f eq 'i';
      $d = $d > 0 ? sprintf "%.2f%%", $d : '' if $f eq 'p';
      
      $d = $r + 1 if $d eq '(ROW)';
      $d = "<span class=hitext>$d</div>" if $oc > 1 and $c == $oc and ! $$rr[0];
      $text .= "<td class=etstd $or[$c]>$d</td>\n";
      }
    $text .= "</tr>\n";
    }
  $text .= "</table>\n";
  return $text;
}

sub table_sum
{
  my $ar = shift;
  my $col = shift;
  my $sum = 0;
  for my $rr ( @$ar )
    {
    next if $rr->[0] ne ''; # skip HEAD,FOOT
    $sum += $rr->[$col];
    }
  return $sum;
}

sub table_ave
{
  my $ar = shift;
  my $col = shift;
  my $sum = 0;
  my $cnt = 0;
  for my $rr ( @$ar )
    {
    next if $rr->[0] ne ''; # skip HEAD,FOOT
    $sum += $rr->[$col];
    $cnt++;
    }
  return $cnt > 0 ? $sum / $cnt : 0;
}

sub table_med
{
  my $ar = shift;
  my $col = shift;
  my @sum;
  for my $rr ( @$ar )
    {
    next if $rr->[0] ne ''; # skip HEAD,FOOT
    push @sum, $rr->[$col];
    }
  @sum = sort { $b <=> $a } @sum;  
  return $sum[ @sum / 2 ];
}

##############################################################################

sub time_fmt
{
  my $data = shift;
  my $d = int($data / (24*60*60));
  my $h = int(($data % (24*60*60)) / (60*60));
  my $m = int(($data % (60*60)) / 60);        
  my $s = $data % 60;                 
  return sprintf "%6dd %02dh %02dm %02ds", $d, $h, $m, $s;
}

sub num_fmt
{
  my $s = shift;
  $s = reverse $s;
  1 while $s =~ s/(\d\d\d)(\d)([^']*)$/$1'$2$3/;
  $s = reverse $s;
}

sub html_save
{
  my $file = shift;
  $file = "$output_dir/$file.html";
  my $text = join '', $html_header, @_, $html_footer;
  my $time = scalar localtime();
  $text =~ s/\[TIME\]/$time/g;
  $text =~ s/\[ENV:([^\]]+)\]/$ENV{$1}/g;
  $text =~ s/href=!/href=$web_prefix/gio;
  $text =~ s/href=#/href=$web_dir/gio;
  $text =~ s/src=#/src=$web_dir/gio;
  $text =~ s/\[T\:([^\]]*)\]/$translation{$1}||$1/gie; # \: is to avoid get_trans.pl
  save_file( $file, $text ) or die "$0: cannot save html file: $file\n";
  #print STDERR "file $file written ok\n";
}

sub color_name
{
  my $s = shift;
  my $cs = $nam_color{ $s } || $s;
  # return $s unless $cs;
  my $c = $cs =~ s/\^(.)/"<font color=#" . $COLOR{ lc $1 } . ">"/gie;
  $cs .= "</font>" x $c;
  return "<span class=name>$cs</name>";
}

sub tr_class
{
  $_[0] % 2 ? 'r1' : 'r2';
}

sub hash_max
{
  my $hr = shift;
  my $max_k = undef;
  my $max_v = undef;
  while( my ( $k, $v ) = each %$hr )
    {
    if( $max_v eq '' or $v > $max_v )
      {
      $max_k = $k;
      $max_v = $v;
      next;
      }
    }
  #print STDERR "$max_k, $max_v\n";
  return ( $max_k, $max_v );
}

sub hash_sum
{
  my $hr = shift;
  my $sum = 0;
  while( my ( $k, $v ) = each %$hr )
    {
    next if $k =~ /^!/;
    $sum += $v;
    }
  return $sum;
}

our %CLAN_CACHE;
sub get_clan
{
  my $s = shift;
  my $cc = $CLAN_CACHE{ $s };
  return $cc if $cc;
  return uc '!' . $clan_map{ $s } if %clan_map;
  my $c = '';
  $c = $1 if $s =~ /^\[([^\]]+)\]/o;
  $c = $1 if $s =~ /\[([^\]]+)\]$/o;
  $c = $1 if $s =~ /^([A-Z]{2,3})[:\-\|]/io;
  return $CLAN_CACHE{ $s } = uc "!$c";
}

sub load_file
{
  my $file = shift;
  my $i;
  open( $i, $file ) or return undef;
  local $/ = undef;
  my $s = <$i>;
  close $i;
  return $s;
}

sub save_file
{
  my $file = shift;
  my $o;
  open( $o, ">$file" ) or return 0;
  print $o @_;
  close $o;
  return 1;
}

###TEST#######################################################################

=pod

Axis   HELLS BELLS       0   1   0  0   0   0     0   125    0      0
                       4 0   1   0  0   0   1    0 0  125    0      0 
                       
Allies ValBo             2   4   0  1  33  17   360   800   66     31                       
                       0 1 133657 0 8 0 0 0 6 9 1 0 0 14 28 1 0 0 0 2 0 0 0 0 4 0 0 0 0 0 0 4 0 360 800 66 19 8 20 3

+++++++++
WS WeaponStats: 2 1 3145756 0 0 0 1 0 3 14 0 0 0 0 0 0 1 0 0 0 0 1 0 3 8 1 0 1 238 472 0 65 9 5                       
Axis   [LUD]CADE   1   3   0  0  25   0   238   472    0     14 
               WS 2 1 
               3145756 (1100000000000000011100)
                         0  0 0 1 0 
                         3 14 0 0 0 
                         0  0 0 1 0 
                         0  0 0 1 0 
                 MP-40   3  8 1 0 1 
                 
                 238 472 0 
                 65 9 5                       
                 
Weapon     Acrcy Hits/Atts Kills Deaths Headshots
-------------------------------------------------
Colt     :                     0      1         0
MP-40    :  21.4    3/14       0      0         0
Thompson :                     0      1         0
Garand   :                     0      1         0
K43 Rifle:  37.5    3/8        1      0         1

Damage Given: 238     Team Damage: 0
Damage Recvd: 472

Rank: Schutze (14 XP)
Skills         Level/Points
---------------------------
Battle Sense   0 (9/20)
Covert Ops     0 (5/20)
                 
------------------------------------------------------------------------------

tova e po-skoro tehnicheska tema taka che ako na nqkoi ne mu e interesno da q 
propusne oshte sega

mai uspqh da razbera kakvo oznachava WeaponStats ot log-a, bih iskal da go 
post-na tuk v sluchai
che na nqkoi mu potrqbva ili mu e interesno

ta v loga nakraq na vsqka igra se poqvqva tova (t.e. podobni redove ):

WeaponStats: 2 1 3145756 0 0 0 1 0 3 14 0 0 0 0 0 0 1 0 0 0 0 1 0 3 8 1 0 1 238 472 0 65 9 5

i sega edno po edno:

2 -- nomer na client-a (player)
1 -- tova mai e za kolko rounds sa tezi danni
3145756 -- bitmap s izpolzvanite oryjiq (binarno e: 1100000000000000011100) t.e. 
           izpolzvani sa 5 oryjiq (colt,mp40,thompson,garand,k43) sledvat grupi 
           ot po 5 chisla za vsqko polzvano oryjie i 
           oznachavat: hits, shots, kills, deaths, headshots

t.e. gornite chisla predstavlqvat tova:

Weapon Acrcy Hits/Atts Kills Deaths Headshots
-------------------------------------------------
Colt : 0 1 0
MP-40 : 21.4 3/14 0 0 0
Thompson : 0 1 0
Garand : 0 1 0
K43 Rifle: 37.5 3/8 1 0 1

sledvashtite 3 chisla sa:

Damage Given: 238
Damage Recvd: 472
Team Damage: 0

sled tova pak ima bitmap na izpolzvanite abilities (clasove)
65 -- binarno e 1000001 i levelite za tezi abilities:

Skills Level/Points
---------------------------
Battle Sense 0 (9/20)
Covert Ops 0 (5/20)

Rank-a mai samo se sybira, t.e. e 14

voila

btw: thanx na Lg-Badluck che mi obyrna vnimanie na komandata /statsdump!

cheers! 


=cut

###EOF########################################################################

